% The well known N-queens problem.

:- use_module(library(lists)).
:- use_module(library(clpfd)).
:- use_module(fdbg).

%% This is only necessary when the constraint is used earlier in the code
%% than it is declared.
:- fdbg_constraint no_threat/3.

nqueens(N, Queens) :-
	bb_put(board_size, N),
	length(Queens, N),
	fdbg_assign_name(Queens, queen),
	domain(Queens, 1, N),
	constrain_all(Queens),
	asymmetric(N, Queens),		% break symmetry
	labeling([ff,enum], Queens).

constrain_all([]).
constrain_all([X|Xs]):-
	constrain_between(X, Xs, 1),
	constrain_all(Xs).

constrain_between(_X, [], _N).
constrain_between(X, [Y|Ys], N) :-
	no_threat(X, Y, N),
	N1 is N+1,
	constrain_between(X, Ys, N1).

% no_threat(X,Y,I) +:
% 	X in \({Y} \/ {Y+I} \/ {Y-I}),
% 	Y in \({X} \/ {X+I} \/ {X-I}).
	
no_threat(X, Y, I) :-
	fd_global(no_threat(X,Y,I), 1, [val(X),val(Y)]).

:- multifile clpfd:dispatch_global/4.
clpfd:dispatch_global(no_threat(X,Y,I), S, S, Actions) :-
	(   integer(X)
	->  no_threat_prop(Y, X, I, Actions)
	;   integer(Y)
	->  no_threat_prop(X, Y, I, Actions)
	;   Actions = []
	).

no_threat_prop(Y, X, I, [Y in_set S, exit]) :-
	XpI is X+I,
	XmI is X-I,
	list_to_fdset([XmI,X,XpI], S0),
	fdset_complement(S0, S).

asymmetric(N, [Q1|Qs]) :-
	last(Qs, Q2),
	Q1 #< Q2           #-- 'vertical reflection symmetry',
	call(Q1 #=< N+1-Q2 #-- 'central reflection symmetry').

:- multifile fdbg:legend_portray/3.
fdbg:legend_portray(Name, Var, After) :-
	write(Name), write(': '),
	print_row(Var, After).

print_row(Var, After) :-
	bb_get(board_size, N),
	fd_set(Var, Now),
	write('['),
	print_fields(1, N, Now, After),
	write(']').

print_fields(I, N, _, _) :-
	I > N, put_code(0' ), !.
print_fields(I, N, Now, After) :-
	put_code(0' ),
	(   fdset_member(I, After)
	->  put_code(0'X)
	;   fdset_member(I, Now)
	->  put_code(0'-)
	;   put_code(0'.)
	),
	I1 is I+1,
	print_fields(I1, N, Now, After).

nqueens_show(Constraint, Actions, _) :-
	fdbg_current_name(Queens0, queen),
	fdbg_annotate(Queens0, Actions, Queens, _),
	print(Constraint), nl,
	print_board(Queens),
	fdbg_transform_actions(Actions, Queens, Actions1),
	(   member(A0, [fail,fail(_,_),exit]),
	    member(A0, Actions1),
	    functor(A0, A, _)
	->  format('    Constraint ~ped.~n', [A])
	;   true
	),
	nl.

print_board([]).
print_board([fdvar(_,Var,After)|Qs]) :- !,
	write('    '),
	print_row(Var, After), nl,
	print_board(Qs).
print_board([V|Qs]) :-
	write('    '),
	fdset_singleton(Set, V),
	print_row(V, Set), nl,
	print_board(Qs).

run_fdbg :-
	fdbg_on([constraint_hook(nqueens_show)]).
